// DIOLAN to DISB Project 1.cpp : This file contains the 'main' function. 
//

/*This cpp-file is meant to provide a demonstration of how the Diolan DLN-4 adapter in conjuction 
with an Ibsen Photonics spectrometer, can be operated. It is up to the individual use to adapt the 
script to their needs.

Running this script will estblish a connection to the DISB hardware, perform read/write to the majority of the 
registers, read calibration coefficients stored on the hardware and save it to a text file (example.txt).
Two spectra will the be taken, one with 10 ms integration time and one with 100 ms intration time. 

More information can be found in the Diolan C/C++ API Manual, which can be downloaded from Diolan's website. 
Ibsen Photonics' spectrometers works via an SPI interface and other interfaces availiable on the Diolan adapter 
boards can not be utilized. 

Communication with a device are performed via messages (Packages of data) which commes in 3 different types.
Commands:	Instructions to the device	
Responses:	After a command has executed a response is returned (Was the command successfull or not? or requested data)
Events:		Contains information about some change that took place. 

Both commands, responses and events are sent through the DLN libary.

[						] -- command -- >	[				] -- command -- >	[ Device 1 ]
[	User Application	] <-- response --	[	DLN libary	] <-- response --	[ Device 2 ]
[						] <-- event --		[				] <-- event --		[ Device N ]
*/


/*Inclusion of header file and dln liberary.
All additional files should be placed in the same folder as this file.
The files can be downloaded from: 
https://diolan.com/dln-4s	--> DOCUMENTATION AND DOWNLOADS --> MS Windows Setup Package
After download and installation the header, dll and libeary files should be available under:
Program Files\Diolan\DLN\examples\c_cpp\bin   &   Program Files\Diolan\DLN\examples\c_cpp\common
*/
#include <iostream>
#include <string>
#include <sstream>
#include <fstream>
#include <chrono>
#include <thread>
#include "dln_generic.h"
#include "dln_spi_master.h"
#include "dln_spi_slave.h"
#include "dln_gpio.h"
//#include "matplotlibcpp.h"

#pragma comment(lib,"dln.lib")
using namespace std;

//Program execution beginsand ends there.
int main(void)
{
	using namespace std::this_thread; // sleep_for, sleep_until
	using namespace std::chrono; // nanoseconds, system_clock, seconds
	/*Establishing a connection to the spectrometer, via the DlnOpenDevice function.
	The DlnOpenDevice function takes 2 inputs.
	First input: An index number of the adapter to open. Useful if several adapters are connected.
	Second input: A pointer to the varaible receiving device handle. The handle is declared before opening the device.
	*/
	typedef uint16_t HDLN;
	HDLN handle;
	DLN_RESULT result = DlnOpenDevice(0, &handle);

	/*
	If the returned result  == DLN_SUCCEEDED, then the a connection to the adapter has successfully been established.
	*/
	if (DLN_SUCCEEDED(result))
		printf("The adapter has succesfully been opened \n");
	else
		printf("Something went wrong and the adapter was not opened \n");


	/*
	Here the Diolan hardware version, serial number and ID are retreived from the adapter.
	The different hardware models are denoted by the follwing codes.
	DLN5	0x0500
	DLN4M	0x0401
	DLN4S	0x0402
	DLN3	0x0300
	DLN2	0x0200
	DLN1	0x0100
	*/
	DLN_VERSION version;
	uint32_t sn, id;
	DlnGetVersion(handle, &version);
	DlnGetDeviceSn(handle, &sn);
	DlnGetDeviceId(handle, &id);

	//Print the hardware type, sn and ID
	printf("Device HwType = 0x%x \nSN = 0x%x \nID = 0x%x\n",
		version.hardwareType, sn, id);


	// Set SPI frequency to 24MHz and print the frequency stored in the Diolan board.
	uint32_t frequency;
	DlnSpiMasterSetFrequency(handle, 0, 24000000, &frequency);
	printf("The frequency of the Diolan controller %d", frequency);


	// Set the Clock polarity and phase 
	DlnSpiMasterSetCpol(handle, 0, 0x0);
	DlnSpiMasterSetCpha(handle, 0, 0x1);
	
	// Check the Tranmission Mode has been set correctly. 
	/* 
	00 -- CPHA = 0, CPOL = 0
	01 -- CPHA = 0, CPOL = 1
	10 -- CPHA = 1, CPOL = 0
	11 -- CPHA = 1, CPOL = 1
	*/
	uint8_t mode;
	DlnSpiMasterGetMode(handle, 0, &mode);
	cout << endl;
	printf("Mode ? 0x%x", mode);

	// Enable SPI master 
	uint16_t conflict;
	DlnSpiMasterEnable(handle, 0, &conflict);

	// Enable the two Chip Select (CS or SS) lines. 
	/*
	These line are enabled when set low so.
	CS0 (SS0) = 1111 1110 (0xFE) - For setting register values.
	CS1 (SS1) = 1111 1101 (0xFD) - For readout of bulk data when acquiring a spectrum
	*/
	DlnSpiMasterSSEnable(handle, 0, 0xfe);
	DlnSpiMasterSSEnable(handle, 0, 0xfd);



	//-------------------------------------------------------------------------------------------------------------------------------
	//-------------------------------------------------------------------------------------------------------------------------------
	/*
	Examples of read/write to the majortiy of different register avaliable on the DISB hardware
	*/
	//-------------------------------------------------------------------------------------------------------------------------------
	//-------------------------------------------------------------------------------------------------------------------------------


	// Set the framesize of the data being transferred.
	// When reading and writing parameters registers 8 bits are used and when reading a spectrum 16 bit values are used.
	DlnSpiMasterSetFrameSize(handle, 0, 8);

	// Setting the currently used Chip select to CS0 (SS0)
	DlnSpiMasterSetSS(handle, 0, 0xfe);

	// Create a .txt file for writing parameters stored on the DISB hardware to
	ofstream Textfile;
	Textfile.open("example.txt");

	// Create and two array of three 8-bit bytes.
	uint8_t input[3], output[3];

	// Read the PCB serial number of register 1. 
	output[0] = 0x06;
	DlnSpiMasterReadWrite(handle, 0, 3, output, input);
	uint16_t combined = (input[1] << 8) | (input[2] & 0xff);
	Textfile << "PCB_SN: ";
	Textfile << combined;

	// Read the Hardware version of register 2. 
	output[0] = 0x0A;
	DlnSpiMasterReadWrite(handle, 0, 3, output, input);
	combined = (input[1] << 8) | (input[2] & 0xff);
	Textfile << "\nHardware version: ";
	Textfile << combined;

	// Read the Firmware version of register 3. 
	output[0] = 0x0E;
	DlnSpiMasterReadWrite(handle, 0, 3, output, input);
	combined = (input[1] << 8) | (input[2] & 0xff);
	Textfile << "\nFirmware version: ";
	Textfile << combined;

	// Read the Detector type of register 4. 
	output[0] = 0x12;
	DlnSpiMasterReadWrite(handle, 0, 3, output, input);
	combined = (input[1] << 8) | (input[2] & 0xff);
	Textfile << "\nDetector type: ";
	Textfile << combined;

	// Read the Pixel per Image of register 5. 
	output[0] = 0x16;
	DlnSpiMasterReadWrite(handle, 0, 3, output, input);
	uint16_t PixelPerImage = (input[1] << 8) | (input[2] & 0xff);
	Textfile << "\nPixel per Image: ";
	Textfile << PixelPerImage;

	// Read the number Calibration Character of register 6. 
	output[0] = 0x1A;
	DlnSpiMasterReadWrite(handle, 0, 3, output, input);
	combined = (input[1] << 8) | (input[2] & 0xff);
	Textfile << "\nNumber of characters of calibration coeffients: ";
	Textfile << combined;

	// Read the trigger delay LSB of register 13. 
	output[0] = 0x36;
	DlnSpiMasterReadWrite(handle, 0, 3, output, input);
	uint16_t combined1 = (input[1] << 8) | (input[2] & 0xff);
	// Read the trigger delay MSB of register 14. 
	output[0] = 0x3A;
	DlnSpiMasterReadWrite(handle, 0, 3, output, input);
	uint16_t combined2 = (input[1] << 8) | (input[2] & 0xff);
	uint32_t combined3 = (combined1 << 16) | (combined2 & 0xffff);

	Textfile << "\nTrigger delay: ";
	Textfile << combined3;

	// Read the ADC Programmable Gain Amplification (PGA) number of register 15. 
	output[0] = 0x3E;
	DlnSpiMasterReadWrite(handle, 0, 3, output, input);
	combined = (input[1] << 8) | (input[2] & 0xff);
	Textfile << "\nADC Programmable Gain Amplification (PGA) number: ";
	Textfile << combined;
	
	// Read the ADC offset number of register 16. 
	output[0] = 0x42;
	DlnSpiMasterReadWrite(handle, 0, 3, output, input);
	combined = (input[1] << 8) | (input[2] & 0xff);
	Textfile << "\nADC offset number: ";
	Textfile << combined;

	// Read the Spectrometer serial number of register 22. 
	output[0] = 0x5A;
	DlnSpiMasterReadWrite(handle, 0, 3, output, input);
	combined = (input[1] << 8) | (input[2] & 0xff);
	Textfile << "\nSpectrometer serial number: ";
	Textfile << combined;

	// Read the First pixel to Read of register 23. 
	output[0] = 0x5E;
	DlnSpiMasterReadWrite(handle, 0, 3, output, input);
	combined = (input[1] << 8) | (input[2] & 0xff);
	Textfile << "\nFirst pixel read: ";
	Textfile << combined;

	// Read the Last pixel to Read of register 24. 
	output[0] = 0x62;
	DlnSpiMasterReadWrite(handle, 0, 3, output, input);
	combined = (input[1] << 8) | (input[2] & 0xff);
	Textfile << "\nLast pixel read: ";
	Textfile << combined;

	Textfile << "\n---------------------------------------------------------------";
	Textfile << "\nSetting integration times and obtaining spectra";
	Textfile << "\n---------------------------------------------------------------";

	//-------------------------------------------------------------------------------------------------------------------------------
	/*
	Examples of setting different integration times and reading spectra
	*/
	//-------------------------------------------------------------------------------------------------------------------------------

	/*
	Setting an integration time of 10 ms.
	Integration time is set in increments of 200 ns, so 10 ms => 50.000 => MSB 0x000 - LBS 0xC350
	*/
	// Setting integration time of Register 10 MSB. 
	output[0] = 0x28;
	output[1] = 0x00;
	output[2] = 0x00;
	DlnSpiMasterReadWrite(handle, 0, 3, output, input);
	// Setting integration time of Register 9 LSB. 
	output[0] = 0x24;
	output[1] = 0xC3;
	output[2] = 0x50;
	DlnSpiMasterReadWrite(handle, 0, 3, output, input);

	// Read the integration time of register 10 MSB. 
	output[0] = 0x2A;
	DlnSpiMasterReadWrite(handle, 0, 3, output, input);
	combined1 = (input[1] << 8) | (input[2] & 0xff);
	// Read the integration time of register 9 MSB. 
	output[0] = 0x26;
	DlnSpiMasterReadWrite(handle, 0, 3, output, input);
	combined2 = (input[1] << 8) | (input[2] & 0xff);
	uint32_t IntegrationTime = (combined1 << 16) | (combined2 & 0xffff);

	Textfile << "\nIntegration time in register in counts: ";
	Textfile << IntegrationTime;
	Textfile << "\nIntegration time in register in milliseconds: ";
	Textfile << IntegrationTime * 0.0002;
	Textfile << " [ms]";

	//-------------------------------------------------------------------------------------------------------------------------------
	// Reading a spectrum

	// Writing to the Sensor Control register 8.
	// Perform soft reset of the image transfer buffer
	output[0] = 0x20;
	output[1] = 0x00;
	output[2] = 0x10;
	DlnSpiMasterReadWrite(handle, 0, 3, output, input);
	// Trigger an exposure
	output[0] = 0x20;
	output[1] = 0x00;
	output[2] = 0x01;
	DlnSpiMasterReadWrite(handle, 0, 3, output, input);

	//Wait 10 ms second for the spectrum to be gathered.
	long long IntegrationTimeinMs;
	IntegrationTimeinMs = IntegrationTime * 0.0002 * 50;
	sleep_for(milliseconds(IntegrationTimeinMs));

	//-------------------------------------------------------------------------------------------------------------------------------
	// Change the chip select from 0 to 1 (SS0 --> SS1) and Framesize from 8-bit bytes to 16-bit
	DlnSpiMasterSetSS(handle, 0, 0xfd);
	DlnSpiMasterSetFrameSize(handle, 0, 16);
	Textfile << "\n";


	uint16_t inputBuffer[PixelPerImage], outputBuffer[PixelPerImage];
	DlnSpiMasterReadWrite16(handle, 0, PixelPerImage, outputBuffer, inputBuffer);
	//Print received data
	for (int i = 0; i < PixelPerImage; i++) {
		Textfile << inputBuffer[i];
		Textfile << ",";
	}
	sleep_for(milliseconds(100));
	//-------------------------------------------------------------------------------------------------------------------------------
	// Change the chip select from 1 to 0 (SS1 --> SS0) and Framesize from 16-bit bytes to 8-bit
	DlnSpiMasterSetSS(handle, 0, 0xfe);
	DlnSpiMasterSetFrameSize(handle, 0, 8);
	Textfile << "\n";

	// Set the integration time to 100 ms
	// Read the integration time of register 10 MSB. 
	output[0] = 0x28;
	output[1] = 0x00;
	output[2] = 0x07;
	DlnSpiMasterReadWrite(handle, 0, 3, output, input);
	combined1 = (input[1] << 8) | (input[2] & 0xff);
	// Read the integration time of register 9 MSB. 
	output[0] = 0x24;
	output[1] = 0xA1;
	output[2] = 0x20;
	DlnSpiMasterReadWrite(handle, 0, 3, output, input);

	// Read the integration time of register 10 MSB. 
	output[0] = 0x2A;
	DlnSpiMasterReadWrite(handle, 0, 3, output, input);
	combined1 = (input[1] << 8) | (input[2] & 0xff);
	// Read the integration time of register 9 MSB. 
	output[0] = 0x26;
	DlnSpiMasterReadWrite(handle, 0, 3, output, input);
	combined2 = (input[1] << 8) | (input[2] & 0xff);
	IntegrationTime = (combined1 << 16) | (combined2 & 0xffff);

	Textfile << "\nIntegration time in register in counts: ";
	Textfile << IntegrationTime;
	Textfile << "\nIntegration time in register in milliseconds: ";
	Textfile << IntegrationTime * 0.0002;
	Textfile << " [ms]";

	//-------------------------------------------------------------------------------------------------------------------------------
	// Trigger exposure

	// Writing to the Sensor Control register 8.
	// Perform soft reset of the image transfer buffer
	output[0] = 0x20;
	output[1] = 0x00;
	output[2] = 0x10;
	DlnSpiMasterReadWrite(handle, 0, 3, output, input);
	// Trigger an exposure
	output[0] = 0x20;
	output[1] = 0x00;
	output[2] = 0x01;
	DlnSpiMasterReadWrite(handle, 0, 3, output, input);

	//Wait 10 ms second for the spectrum to be gathered.
	IntegrationTimeinMs = IntegrationTime * 0.0002 * 10;
	sleep_for(milliseconds(IntegrationTimeinMs));

	//-------------------------------------------------------------------------------------------------------------------------------
	// Change the chip select from 0 to 1 (SS0 --> SS1) and Framesize from 8-bit bytes to 16-bit
	DlnSpiMasterSetSS(handle, 0, 0xfd);
	DlnSpiMasterSetFrameSize(handle, 0, 16);
	Textfile << "\n";

	DlnSpiMasterReadWrite16(handle, 0, PixelPerImage, outputBuffer, inputBuffer);
	//Print received data
	for (int i = 0; i < PixelPerImage; i++) {
		Textfile << inputBuffer[i];
		Textfile << ",";
	}
	
	//-------------------------------------------------------------------------------------------------------------------------------
	DlnSpiMasterSetSS(handle, 0, 0xfe);
	DlnSpiMasterSetFrameSize(handle, 0, 8);
	Textfile << "\n";


	Textfile.close();


	
	//Closing of the adapter connection.
	DlnCloseHandle(handle);
}


